"""Vertex IDs of keypoints for various models."""
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de


vertex_ids = {
    'smal': {
        'lnose':	    1845,
        'reye':		    3044,
        'leye':		    1086,
        'rear_tip':		2508,
        'lear_tip':		543,
        'jaw':		    11,
        'tail':		    1836,
    },
    'hsmal': {
        'lnostril':		575,
        'reye':		    1191,
        'leye':		    378,
        'rear_tip':		1039,
        'lear_tip':		226,
        'jaw':		    611,
        'tail':		    752,
    },
    # NEED UPDATING. CURRENTLY SAME AS VAREN
    'hsmal+': {
        'nose':		    1738,
        'reye':		    3498,
        'leye':		    1596,
        'rear_tip':		2769,
        'lear_tip':		853,
        'jaw':		    1732,
        'tail':		    1910,
    },
    'varen': {
        'lfront_foot_tip': 1473,
        'rfront_foot_tip': 3390,
        'lfront_sesamoid': 1505,
        'rfront_sesamoid': 3422,
        'lback_foot_tip': 1410,
        'rback_foot_tip': 3327,
        'lback_calcaneous': 744,
        'rback_calcaneous': 2661,
        'back_left_foot_back': 1359,
        'back_right_foot_back': 3276,
        'front_left_foot_back': 1446,
        'front_right_foot_back': 3363,
        'lfront_knee': 663,
        'rfront_knee': 2580,
        'lear_tip': 852,
        'rear_tip': 2769,
        'tail_start': 1717,
        'tail_tip': 1915,
        'half_tail': 1899,
        'lear_base_back': 798,
        'rear_base_back': 2714,
        'lear_base_front_internal': 881,  # 884,
        'lear_base_front_external': 904,  # 1314,
        'rear_base_front_internal': 2798,  # 2801,
        'rear_base_front_external': 2821,  # 3231,
        'between_nostrils_base': 1739,
        'lnostril': 1154,
        'rnostril': 3071,  # 3071,
        'leye': 1250,  # 1583,
        'reye': 3498,
        'rcalcaneous': 2661,
        'lcalcaneous': 744,
        'lfront_coffin': 1473,
        'rfront_coffin': 3390,
        'lback_coffin': 1410,
        'rback_coffin': 3327,
        'end_sacral_vertebrae': 1717,
        'rcarpus_back': 2005,
        'lcarpus_back': 88,
        'tip_of_head': 1691,
        'rposterior': 2536,  # 2176,
        'lposterior': 619,  # 259,
        'middle_back': 1712

        }
    # 'varen': {
    #     'nose':		    1738,
    #     'reye':		    3498,
    #     'leye':		    1596,
    #     'rear':		    2769,
    #     'lear':		    853,
    #     'jaw':		    1732,
    #     'tail':		    1910,
    # }
}
